package meerkat.bulletinboard;

import com.google.common.util.concurrent.FutureCallback;
import com.google.common.util.concurrent.Futures;
import com.google.common.util.concurrent.ListeningScheduledExecutorService;
import com.google.common.util.concurrent.MoreExecutors;
import com.google.protobuf.ByteString;
import meerkat.bulletinboard.workers.SingleServerGetRedundancyWorker;
import meerkat.bulletinboard.workers.SingleServerPostMessageWorker;
import meerkat.bulletinboard.workers.SingleServerReadMessagesWorker;
import meerkat.protobuf.BulletinBoardAPI.*;
import meerkat.protobuf.Voting.BulletinBoardClientParams;

import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

/**
 * Created by Arbel Deutsch Peled on 28-Dec-15.
 *
 * This class implements the asynchronous Bulletin Board Client interface
 * It only handles a single Bulletin Board Server
 * If the list of servers contains more than one server: the server actually used is the first one
 * The class further implements a delayed access to the server after a communication error occurs
 */
public class SingleServerBulletinBoardClient extends SimpleBulletinBoardClient implements AsyncBulletinBoardClient {

    private final int MAX_RETRIES = 10;

    protected ListeningScheduledExecutorService executorService;

    private long lastServerErrorTime;

    protected final long failDelayInMilliseconds;

    /**
     * Notify the client that a job has failed
     * This makes new scheduled jobs be scheduled for a later time (after the given delay)
     */
    protected void fail() {

        // Update last fail time
        lastServerErrorTime = System.currentTimeMillis();

    }

    /**
     * This method adds a worker to the scheduled queue of the threadpool
     * If the server is in an accessible state: the job is submitted for immediate handling
     * If the server is not accessible: the job is scheduled for a later time
     * @param worker is the worker that should be scheduled for work
     * @param callback is the class containing callbacks for handling job completion/failure
     */
    protected void scheduleWorker(SingleServerWorker worker, FutureCallback callback){

        long timeSinceLastServerError = System.currentTimeMillis() - lastServerErrorTime;

        if (timeSinceLastServerError >= failDelayInMilliseconds) {

            // Schedule for immediate processing
            Futures.addCallback(executorService.submit(worker), callback);

        } else {

            // Schedule for processing immediately following delay expiry
            Futures.addCallback(executorService.schedule(
                    worker,
                    failDelayInMilliseconds - timeSinceLastServerError,
                    TimeUnit.MILLISECONDS),
                    callback);

        }

    }

    /**
     * Inner class for handling simple operation results and retrying if needed
     */
    class RetryCallback<T> implements FutureCallback<T> {

        private SingleServerWorker worker;
        private ClientCallback<T> clientCallback;

        public RetryCallback(SingleServerWorker worker, ClientCallback<T> clientCallback) {
            this.worker = worker;
            this.clientCallback = clientCallback;
        }

        @Override
        public void onSuccess(T result) {
            clientCallback.handleCallback(result);
        }

        @Override
        public void onFailure(Throwable t) {

            // Notify client about failure
            fail();

            // Check if another attempt should be made

            worker.decMaxRetry();

            if (worker.isRetry()) {
                // Perform another attempt
                scheduleWorker(worker, this);
            } else {
                // No more retries: notify caller about failure
                clientCallback.handleFailure(t);
            }

        }

    }


    public SingleServerBulletinBoardClient(int threadPoolSize, long failDelayInMilliseconds) {

        executorService = MoreExecutors.listeningDecorator(Executors.newScheduledThreadPool(threadPoolSize));

        this.failDelayInMilliseconds = failDelayInMilliseconds;

        // Set server error time to a time sufficiently in the past to make new jobs go through
        lastServerErrorTime = System.currentTimeMillis() - failDelayInMilliseconds;

    }

    /**
     * Stores database location, initializes the web Client and
     * @param clientParams contains the data needed to access the DBs
     */
    @Override
    public void init(BulletinBoardClientParams clientParams) {

        // Perform usual setup
        super.init(clientParams);

        // Remove all but first DB address
        String dbAddress = meerkatDBs.get(0);
        meerkatDBs = new LinkedList<String>();
        meerkatDBs.add(dbAddress);

    }

    @Override
    public MessageID postMessage(BulletinBoardMessage msg, ClientCallback<Boolean> callback) {

        // Create worker with redundancy 1 and MAX_RETRIES retries
        SingleServerPostMessageWorker worker = new SingleServerPostMessageWorker(meerkatDBs.get(0), msg, MAX_RETRIES);

        // Submit worker and create callback
        scheduleWorker(worker, new RetryCallback(worker, callback));

        // Calculate the correct message ID and return it
        digest.reset();
        digest.update(msg.getMsg());
        return MessageID.newBuilder().setID(ByteString.copyFrom(digest.digest())).build();

    }

    @Override
    public MessageID postBatch(CompleteBatch completeBatch, ClientCallback<Boolean> callback) {
        return null;
    }

    @Override
    public void beginBatch(byte[] signerId, int batchId, List<String> tagList, ClientCallback<Boolean> callback) {

    }

    @Override
    public void postBatchData(byte[] signerId, int batchId, List<BatchData> batchDataList,
                              int startPosition, ClientCallback<Boolean> callback) {

    }

    @Override
    public void postBatchData(byte[] signerId, int batchId, List<BatchData> batchDataList, ClientCallback<Boolean> callback) {

    }

    @Override
    public void closeBatch(CloseBatchMessage closeBatchMessage, ClientCallback<?> callback) {

    }

    @Override
    public void getRedundancy(MessageID id, ClientCallback<Float> callback) {

        // Create worker with no retries
        SingleServerGetRedundancyWorker worker = new SingleServerGetRedundancyWorker(meerkatDBs.get(0), id, 1);

        // Submit job and create callback
        scheduleWorker(worker, new RetryCallback(worker, callback));

    }

    @Override
    public void readMessages(MessageFilterList filterList, ClientCallback<List<BulletinBoardMessage>> callback) {

        // Create job with no retries
        SingleServerReadMessagesWorker worker = new SingleServerReadMessagesWorker(meerkatDBs.get(0), filterList, 1);

        // Submit job and create callback
        scheduleWorker(worker, new RetryCallback(worker, callback));

    }

    @Override
    public void readBatch(byte[] signerId, int batchId, ClientCallback<CompleteBatch> callback) {

    }

    @Override
    public void subscribe(MessageFilterList filterList, MessageHandler messageHandler) {

    }

    @Override
    public void close() {

        super.close();

        executorService.shutdown();

    }
}
