package meerkat.bulletinboard;

import com.google.protobuf.ByteString;

import meerkat.bulletinboard.workers.MultiServerGetRedundancyWorker;
import meerkat.bulletinboard.workers.MultiServerPostMessageWorker;
import meerkat.bulletinboard.workers.MultiServerReadMessagesWorker;
import meerkat.comm.CommunicationException;
import meerkat.protobuf.BulletinBoardAPI.*;
import meerkat.protobuf.Voting.*;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

/**
 * Created by Arbel Deutsch Peled on 05-Dec-15.
 * Thread-based implementation of a Async Bulletin Board Client.
 * Features:
 * 1. Handles tasks concurrently.
 * 2. Retries submitting
 */
public class ThreadedBulletinBoardClient extends SimpleBulletinBoardClient implements AsyncBulletinBoardClient {

    // Executor service for handling jobs
    private final static int JOBS_THREAD_NUM = 5;
    ExecutorService executorService;

    // Per-server clients
    List<SingleServerBulletinBoardClient> clients;

    private final static int POST_MESSAGE_RETRY_NUM = 3;
    private final static int READ_MESSAGES_RETRY_NUM = 1;
    private final static int GET_REDUNDANCY_RETRY_NUM = 1;

    private static final int SERVER_THREADPOOL_SIZE = 5;
    private static final long FAIL_DELAY = 5000;

    private int minAbsoluteRedundancy;

    /**
     * Stores database locations and initializes the web Client
     * Stores the required minimum redundancy.
     * Starts the Thread Pool.
     * @param clientParams contains the required information
     */
    @Override
    public void init(BulletinBoardClientParams clientParams) {

        super.init(clientParams);

        minAbsoluteRedundancy = (int) (clientParams.getMinRedundancy() * clientParams.getBulletinBoardAddressCount());

        executorService = Executors.newFixedThreadPool(JOBS_THREAD_NUM);

        clients = new ArrayList<SingleServerBulletinBoardClient>(clientParams.getBulletinBoardAddressCount());
        for (String address : clientParams.getBulletinBoardAddressList()){
            SingleServerBulletinBoardClient client = new SingleServerBulletinBoardClient(SERVER_THREADPOOL_SIZE, FAIL_DELAY);
            client.init(BulletinBoardClientParams.newBuilder()
                            .addBulletinBoardAddress(address)
                            .build());
            clients.add(client);
        }

    }

    /**
     * Post message to all DBs
     * Retry failed DBs
     * @param msg is the message,
     * @return the message ID for later retrieval
     * @throws CommunicationException
     */
    @Override
    public MessageID postMessage(BulletinBoardMessage msg, ClientCallback<Boolean> callback){

        // Create job
        MultiServerPostMessageWorker worker =
                new MultiServerPostMessageWorker(clients, minAbsoluteRedundancy, msg, POST_MESSAGE_RETRY_NUM, callback);

        // Submit job
        executorService.submit(worker);

        // Calculate the correct message ID and return it
        digest.reset();
        digest.update(msg.getMsg());
        return MessageID.newBuilder().setID(ByteString.copyFrom(digest.digest())).build();

    }

    @Override
    public MessageID postBatch(CompleteBatch completeBatch, ClientCallback<Boolean> callback) {

        return null;    // TODO: write this

    }

    @Override
    public void beginBatch(byte[] signerId, int batchId, List<String> tagList, ClientCallback<Boolean> callback) {
        // TODO: write this
    }

    @Override
    public void postBatchData(byte[] signerId, int batchId, List<BatchData> batchDataList,
                                   int startPosition, ClientCallback<Boolean> callback) {

        // TODO: write this

    }

    @Override
    public void postBatchData(byte[] signerId, int batchId, List<BatchData> batchDataList, ClientCallback<Boolean> callback) {

        postBatchData(signerId, batchId, batchDataList, 0, callback);   // Write batch from beginning

    }

    @Override
    public void closeBatch(CloseBatchMessage closeBatchMessage, ClientCallback<?> callback) {
        // TODO: write this
    }

    /**
     * Access each database and search for a given message ID
     * Return the number of databases in which the message was found
     * Only try once per DB
     * Ignore communication exceptions in specific databases
     */
    @Override
    public void getRedundancy(MessageID id, ClientCallback<Float> callback) {

        // Create job
        MultiServerGetRedundancyWorker worker =
                new MultiServerGetRedundancyWorker(clients, minAbsoluteRedundancy, id, GET_REDUNDANCY_RETRY_NUM, callback);

        // Submit job
        executorService.submit(worker);

    }

    /**
     * Go through the DBs and try to retrieve messages according to the specified filter
     * If at the operation is successful for some DB: return the results and stop iterating
     * If no operation is successful: return null (NOT blank list)
     */
    @Override
    public void readMessages(MessageFilterList filterList, ClientCallback<List<BulletinBoardMessage>> callback) {

        // Create job
        MultiServerReadMessagesWorker worker =
                new MultiServerReadMessagesWorker(clients, minAbsoluteRedundancy, filterList, READ_MESSAGES_RETRY_NUM, callback);

        // Submit job
        executorService.submit(worker);

    }

    @Override
    public void readBatch(byte[] signerId, int batchId, ClientCallback<CompleteBatch> callback) {
        // TODO: Implement
    }

    @Override
    public void subscribe(MessageFilterList filterList, MessageHandler messageHandler) {
        // TODO: Implement
    }

    @Override
    public void close() {
        super.close();

        try {
            executorService.shutdown();
            while (! executorService.isShutdown()) {
                executorService.awaitTermination(10, TimeUnit.SECONDS);
            }
        } catch (InterruptedException e) {
            System.err.println(e.getCause() + " " + e.getMessage());
        }
    }

}
